// $Id: PackageTest.java,v 1.7 2004/06/14 20:52:16 m31 Exp $
//
// This file is part of the program suite "Principia Mathematica II"
// which is a working prototype for the main project:
// "Hilbert II" - http://www.qedeq.org
//
// Copyright 2001-2004  Michael Meyling <mime@qedeq.org>.
//
// "Principia Mathematica II" is free software; you can redistribute
// it and/or modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.

package com.meyling.principia.logic.basic;


import java.io.File;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.PrintStream;

import com.meyling.principia.argument.Argument;
import com.meyling.principia.argument.ArgumentException;
import com.meyling.principia.io.Output;
import com.meyling.principia.io.ParsingException;
import com.meyling.principia.io.ParsingTable;
import com.meyling.principia.io.TextFileReader;
import com.meyling.principia.io.TextInput;
import com.meyling.principia.io.Utility;


/**
 * General package test.
 *
 * @version $Revision: 1.7 $
 * @author  Michael Meyling
 */
public class PackageTest {


    /** path to test data */
    public static final String FILE_PATH
        = "com/meyling/principia/logic/basic/test/";

    /** file name prefix */
    public static final String TEST_FILE_PREFIX
        = "test.";

    /** input file name postfix */
    public static final String INPUT_FILE_POSTFIX
        = ".input";

    /** output file name postfix */
    public static final String ACTUAL_OUTPUT_FILE_POSTFIX
        = ".output";

    /** expected output file name postfix */
    public static final String EXPECTED_OUTPUT_FILE_POSTFIX
        = ".output.expected";

    /** test input file  */
    public static final String EXCEPTION_TEST_FILE_PREFIX
        = "exception.test.";

    /** seperates several arguments */
    public static final String NEXT = "next";

    /** seperates result from arguments */
    public static final String RESULT = "result";

    /** marks end of result */
    public static final String END = "end";

    /** true */
    public static final String TRUE = "true";

    /** false */
    public static final String FALSE = "false";

    /** creation failed */
    public static final String FAIL = "fail";

    /** output stream */
    public static PrintStream out = System.out;


    /**
     * Runs the following test numbers (or all tests).
     * The result ist printed to <code>System.out</code>.
     *
     * @param   argument    (optional) list of test numbers,
     *                      a leading "-t " runs the exception test
     */
    public static final void main (String[] argument) {
        System.out.println("Testing package com.meyling.principia.logic.basic...");
        System.out.println();
        boolean result = true;
        if (argument.length > 0) {
            if (argument[0].equals("-t")) {
                if (argument.length == 1) {
                    result = testException();
                } else {
                    for (int i = 1; i < argument.length; i++) {
                        result &= testExceptionArgument(argument[i]);
                    }
                }
            } else {
                for (int i = 0; i < argument.length; i++) {
                    result &= testArgument(argument[i]);
                }
            }
        } else {
            result = test();
        }
        System.out.println();
        System.out.println("..." + (result ? "passed!" : "failed!"));
    }


    /**
     * Run all normal tests in directory {@link #FILE_PATH}.
     * All files that start with {@link #TEST_FILE_PREFIX} and
     * end with {@link #INPUT_FILE_POSTFIX} are put into
     * {@link #testArgument}. Iff all these tests return
     * <code>true</code> this method returns true.
     *
     * @return  <code>true</code> iff all tests were ok
     */
    public static final boolean test() {
        boolean result = true;

        String[] testFiles = null;
        final File testDir = new File(FILE_PATH);
        if (testDir.isDirectory()) {
            testFiles = testDir.list(new FilenameFilter() {
                public boolean accept(File dir, String name) {
                    return name.startsWith(TEST_FILE_PREFIX)
                        && name.endsWith(INPUT_FILE_POSTFIX);
                }
            });
            if (testFiles == null) {
                System.out.println("No test files found in directory "
                    + testDir.toString());
                return false;
            } else {
                for (int i = 0; i < testFiles.length; i++) {
                    testFiles[i] = testFiles[i].substring(
                        TEST_FILE_PREFIX.length(),
                        testFiles[i].length() - INPUT_FILE_POSTFIX.length());
                }
            }
        } else {
            System.out.println("Directory " + FILE_PATH + " not found");
        }
        for (int i = 0; i < testFiles.length; i++) {
            result &= testArgument(testFiles[i]);
        }
        return result;
    }


    /**
     * Run all exception tests in directory {@link #FILE_PATH}.
     * All files that start with {@link #EXCEPTION_TEST_FILE_PREFIX} and
     * end with {@link #INPUT_FILE_POSTFIX} are put into
     * {@link #testArgument}. Iff all these tests return
     * <code>true</code> this method returns true.
     *
     * @return  <code>true</code> iff all tests were ok
     */
    public static final boolean testException() {
        boolean result = true;

        String[] testFiles = null;
        final File testDir = new File(FILE_PATH);
        if (testDir.isDirectory()) {
            testFiles = testDir.list(new FilenameFilter() {
                public boolean accept(File dir, String name) {
                    return name.startsWith(EXCEPTION_TEST_FILE_PREFIX)
                        && name.endsWith(INPUT_FILE_POSTFIX);
                }
            });
            if (testFiles == null) {
                System.out.println("No test files found in directory "
                    + testDir.toString());
                return false;
            } else {
                for (int i = 0; i < testFiles.length; i++) {
                    testFiles[i] = testFiles[i].substring(
                        EXCEPTION_TEST_FILE_PREFIX.length(),
                        testFiles[i].length() - INPUT_FILE_POSTFIX.length());
                }
            }
        } else {
            System.out.println("Directory " + FILE_PATH + " not found");
        }
        for (int i = 0; i < testFiles.length; i++) {
            result &= testExceptionArgument(testFiles[i]);
        }
        return result;
    }


    /**
     * Test a single normal test file. The actual file name is build
     * by connecting {@link #FILE_PATH} {@link #TEST_FILE_PREFIX}
     * <code>argument</code> {@link #INPUT_FILE_POSTFIX}.
     * Start and result information is printed to <code>System.out</code>.
     *
     * @param   argument    part of the file name
     * @return  was test ok?
     */
    public static final boolean testArgument(final String argument) {
        final String fileName = FILE_PATH + TEST_FILE_PREFIX
                + argument + INPUT_FILE_POSTFIX;
        System.out.println("\ttesting file " + fileName);
        final int errors = testFile(fileName);
        if (errors > 0) {
            System.out.println("\t" + errors + " errors");
        }
        return errors == 0;
    }


    /**
     * Test a single exception test file. The actual file name is build
     * by connecting {@link #FILE_PATH} {@link #EXCEPTION_TEST_FILE_PREFIX}
     * <code>argument</code> {@link #INPUT_FILE_POSTFIX}.
     * Start and result information is printed to <code>System.out</code>.
     *
     * @param   argument    part of the file name
     * @return  was test ok?
     */
    public static final boolean testExceptionArgument(final String argument) {
        boolean result = false;
        final String fileName = FILE_PATH + EXCEPTION_TEST_FILE_PREFIX
            + argument;
        try {
            System.out.println("\ttesting file " + fileName + INPUT_FILE_POSTFIX);
            out = new PrintStream(new FileOutputStream(
                fileName  + ACTUAL_OUTPUT_FILE_POSTFIX));
            testFile(fileName + INPUT_FILE_POSTFIX);
            out.close();
            final String output = Utility.loadFile(fileName
                + ACTUAL_OUTPUT_FILE_POSTFIX);
            if (!output.equals(Utility.loadFile(fileName
                + EXPECTED_OUTPUT_FILE_POSTFIX))) {
                System.out.println("\t\tis not equal to excpected");
                result = false;
            } else {
                result = true;
            }
        } catch (IOException e) {
            System.out.println(e.toString());
        } catch (Exception e) {
            e.printStackTrace();
            result = false;
        }
        return result;
    }


    /**
     * Test this package with a file.
     * Following tests are implemented:
     * <ul>
     * <li>create</li>
     * <li>equals</li>
     * <li>replace1</li>
     * </ul>
     *
     * @param   fileName    name of file with test data
     * @return  number of errors that occured
     */
    public static final int testFile(final String fileName) {
        int errors = 0;
        TestReader reader = null;
        try {
            reader = new TestReader(fileName);
            String line;
            int startLine;
            line = reader.readLine();
            while (line != null) {
                startLine = reader.getLineNumber();
                try {
                    if (line.startsWith("create")) {
                        if (!testCreate(reader, startLine)) {
                            errors++;
                        }
                    } else if (line.startsWith("equals")) {
                        if (!testEquals(reader, startLine)) {
                            errors++;
                        }

                    } else if (line.startsWith("replace1")) {
                        if (!testReplace1(reader, startLine)) {
                            errors++;
                        }

                    } else {
                        throw new IllegalArgumentException(
                            "Unknown command in line "
                            + reader.getLineNumber() + ":\n"
                            + line);
                    }
                } catch (IllegalArgumentException e) {
                    errors++;
                    if (out == System.out) {    // Q & D
                        e.printStackTrace(out);
                    } else {
                        out.println(e.toString());
                        out.println();
                    }
                }

                line = reader.readLine();
            }
            reader.close();
        } catch (IOException e) {
            System.out.println(e.toString());
            return 1;
        } finally {
            if (reader != null) {
                reader.close();
            }
        }
        return errors;
    }


    /**
     * Test creation of argument.
     *
     * @param   reader       source to read from
     * @param   startLine    source line where test started
     * @return  result of test
     * @throws  IOException  if any file error occured
     */
    public static boolean testCreate(TestReader reader, int startLine)
            throws IOException, IllegalArgumentException {

        boolean creationFailed = true;
        IllegalArgumentException exception = null;
        try {
            // read first (and only) argument
            reader.readArgument(RESULT);
            creationFailed = false;
        } catch (IllegalArgumentException e) {
            exception = e;
        }
        final String argumentString = reader.getArgument();

        // read result
        final boolean result = reader.readBoolean(END);

        if (creationFailed && result) {
            if (out == System.out) {    // Q & D
                exception.printStackTrace();
            }
            out.println("failed line " + startLine
                + ", creation expected:");
            out.println(exception.toString());
            out.println();
            return false;
        } else if (!creationFailed && !result) {
            out.println("failed line " + startLine
                + ", creation should have failed:");
            out.println(argumentString);
            out.println();
            return false;
        } else {
            return true;
        }

    }


    /**
     * Test equality of arguments.
     *
     * @param   reader       source to read from
     * @param   startLine    source line where test started
     * @return  result of test
     * @throws  IOException  if any file error occured
     * @throws  IllegalArgumentException    if an expected creation
     *          failed
     */
    public static boolean testEquals(TestReader reader, int startLine)
            throws IOException, IllegalArgumentException {

        // read first argument
        final Argument argument1 = reader.readArgument(NEXT);

        // read second argument
        final Argument argument2 = reader.readArgument(RESULT);

        // read result
        final boolean result = reader.readBoolean(END);


        if (result) {
            if (!argument1.equals(argument2)
                    || !argument2.equals(argument1)) {
                out.println(argument1);
                out.println(argument2);
                out.println("failed line " + startLine
                    + ", should be equal");
                out.println();
                return false;
            } else {
                return true;
            }
        } else {
            if (argument1.equals(argument2)
                    || argument2.equals(argument1)) {
                out.println(argument1);
                out.println(argument2);
                out.println("failed line " + startLine
                    + ", shouldn't be equal");
                out.println();
                return false;
            } else {
                return true;
            }
        }

    }


    /**
     * Test simple replacement of arguments.
     *
     * @param   reader       source to read from
     * @param   startLine    source line where test started
     * @return  result of test
     * @throws  IOException  if any file error occured
     * @throws  IllegalArgumentException    if an expected creation
     *          failed
     */
    public static boolean testReplace1(TestReader reader, int startLine)
            throws IOException, IllegalArgumentException {

        // read first argument
        final Argument argument1 = reader.readArgument(NEXT);

        // read second argument
        final Argument argument2 = reader.readArgument(NEXT);

        // read third argument
        final Argument argument3 = reader.readArgument(RESULT);


        // read result argument
        final Argument result = reader.readArgumentOrFail(END);

        final Argument actualResult;
        if (result == null) {
            try {
                actualResult = argument1.replace(argument2, argument3);
                out.println(BasicCreator.writeArgument(result));
                out.println("failed line " + startLine
                    + ", no creation expected");
                return false;
            } catch (Exception e) {
                return true;
            }
        } else {
            try {
                actualResult = argument1.replace(argument2, argument3);
                if (!actualResult.equals(result)
                        || !result.equals(actualResult)) {
                    out.println(BasicCreator.writeArgument(actualResult));
                    out.println(BasicCreator.writeArgument(result));
                    out.println("failed line " + startLine
                        + ", actual result and expected result should "
                        + "be equal");
                    out.println();
                    return false;
                } else {
                    return true;
                }
            } catch (ArgumentException e) {
                if (out == System.out) {    // Q & D
                    e.printStackTrace();
                }
                out.println(e.toString());
                out.println("failed line " + startLine
                    + ", creation of result expected");
                out.println();
                return false;
            } catch (IllegalArgumentException e) {
                if (out == System.out) {    // Q & D
                    e.printStackTrace();
                }
                out.println(e.toString());
                out.println("failed line " + startLine
                    + ", creation of result expected");
                out.println();
                return false;
            }
        }
    }


    /**
     * Check if <code>text</code> is equal to {@link #NEXT}.
     *
     * @param   text    compare text
     * @return  result of equal
     */
    private static boolean isNextSeparator(String text) {
        return text.equals(NEXT);
    }


    /**
     * Check if <code>text</code> is equal to {@link #RESULT}.
     *
     * @param   text    compare text
     * @return  result of equal
     */
    private static boolean isResultSeparator(String text) {
        return text.equals(RESULT);
    }


    /**
     * Check if <code>text</code> is equal to {@link #END}.
     *
     * @param   text    compare text
     * @return  result of equal
     */
    private static boolean isEndSeparator(String text) {
        return text.equals(END);
    }


    /**
     * Check if <code>text</code> is equal to any separator:
     * {@link #NEXT}, {@link #RESULT} or  {@link #END}.
     *
     * @param   text    compare text
     * @return  is <code>text</code> equal to any separator
     */
    private static boolean isSeparator(String text) {
        return isNextSeparator(text) || isResultSeparator(text)
            || isEndSeparator(text);
    }


    /**
     * Performes simple argument tests:
     * <ul>
     *   <li>is argument equal to itself</li>
     *   <li>is argument equal to argument.copy() (both directions)</li>
     *   <li>is argument equal to same argument created of
     *        <code>list</code> with
     *        {@link BasicCreator#readArgument} (both
     *        directions)</li>
     *   <li>is argument equal to same argument created of result of
     *        {@link BasicCreator#writeArgument}
     *        with {@link BasicCreator#readArgument}
     *        (both directions)</li>
     *
     * </ul>
     *
     * @param   argument    argument to test
     * @param   list        from this data the argument was created
     * @throws  IllegalArgumentException    if test failed
     */
    public static void testArgument(final Argument argument,
            final StringBuffer list) throws IllegalArgumentException {

        if (!argument.equals(argument)) {
            throw new IllegalArgumentException("no self equality:\n"
                + BasicCreator.writeArgument(argument));
        }
        if (!argument.equals(argument.copy())) {
            throw new IllegalArgumentException("not equal to copy:\n"
                + BasicCreator.writeArgument(argument));
        }
        if (!argument.copy().equals(argument)) {
            throw new IllegalArgumentException("copy not equal:\n"
                + BasicCreator.writeArgument(argument));
        }
        if (!argument.copy().equals(argument.copy())) {
            throw new IllegalArgumentException("copies are not equal:\n"
                + BasicCreator.writeArgument(argument));
        }
        try {
            final StringBuffer out = new StringBuffer();
            BasicCreator.writeArgument(new Output(out), argument);
            final Argument compare = BasicCreator.readArgument(
                    new TextInput(out, "", ""), new ParsingTable());
            if (!argument.equals(compare)
                    || !compare.equals(argument)) {

                throw new IllegalArgumentException("new creation from written"
                    + " argument is not equal to original one:\n"
                    + BasicCreator.writeArgument(argument));
            }
            final StringBuffer out2 = new StringBuffer();
            BasicCreator.writeArgument(new Output(out2), argument);
            if (!out.toString().equals(out2.toString())) {
                throw new IllegalArgumentException("new creation from written"
                    + " argument is not equal to second creation from written"
                    + " one:\n"
                    + BasicCreator.writeArgument(argument));
            }
            final Argument compare2 = BasicCreator.readArgument(
                    new TextInput(list, "", ""), new ParsingTable());
            if (!argument.equals(compare2)
                    || !compare2.equals(argument)) {
                throw new IllegalArgumentException("argument is not equal to"
                    + " recreated argument form written one\n"
                    + BasicCreator.writeArgument(argument));
            }
        } catch (IllegalArgumentException e) {
            throw e;
        } catch (Exception e) {
            throw new IllegalArgumentException("writing argument or creation"
                + "from written argument failed\n" + e.getMessage() );
        }
    }




    /**
     * Special reader for test files with convenient methods for
     * reading arguments.
     *
     * @version $Revision: 1.7 $
     * @author  Michael Meyling
     */
    private static class TestReader extends TextFileReader {


        /** buffer for argument */
        private StringBuffer argumentBuffer = new StringBuffer();

        /** remember starting line number */
        private int argumentLineNumber = 0;


        /**
         * Creates a <code>TestReader</code> object.
         *
         * @param   inputName   name of the input file
         * @throws  IOException if the buffered reader couldn't be created
         */
        private TestReader(String inputName) throws IOException {
            super(inputName);
        }


        /**
         * Read an {@link Argument} and check separator.
         * Lines are read till {@link #isSeparator} returns
         * <code>true</code>. These lines should build an
         * {@link Argument}.
         *
         * @param   separator   this text must follow the argument
         * @return  created argument
         * @throws  IOException if any file error occured
         * @throws  IllegalArgumentException if argument creation
         *          failed, the separator wasn't found or the
         *          file ended unexpected
         */
        private final Argument readArgument(final String separator)
                throws IOException, IllegalArgumentException {
            argumentLineNumber = getLineNumber();
            argumentBuffer.setLength(0);
            String line;
            while ((line = readLine()) != null
                    && !isSeparator(line)) {
                argumentBuffer.append(line).append("\n");
            }
            checkSeparator(line, separator);
            final Argument argument;
            final TextInput in = new TextInput(argumentBuffer, "", "");
            try {
                argument = BasicCreator.readArgument(in, new ParsingTable());
            } catch (ParsingException e) {
/*
                System.out.println("parsing exception");
                System.out.println(e.getDescription());
                System.out.println("Zeile:  " + e.getRow());
                System.out.println("Spalte: " + e.getColumn());
                System.out.println("Line:   " + e.getLine());
                System.out.println(in.showLinePosition());
*/
                final StringBuffer msg = new StringBuffer();
                msg
//                  .append(
//                    (new File(e.getAddress())
//                    .getAbsolutePath()).toString())
//                    .append(":")
                    .append(e.getRow()).append(":").append(e.getColumn())
                    .append(":\n");
                msg.append(e.getDescription()).append("\n");
                msg.append(e.getLine()).append("\n");
                for (int i = 0; i < e.getColumn() - 1; i++) {
                    msg.append(' ');   // Q & D
                }
                msg.append("^");
//                System.out.println(msg);
                throw new IllegalArgumentException(msg.toString());
            } catch (IllegalArgumentException e) {
                throw new IllegalArgumentException(e.getMessage() + "\n"
                    + in.showLinePosition()
                    + "\nin or after line "
                    + argumentLineNumber);
            }
            testArgument(argument, argumentBuffer);
            return argument;
        }


        /**
         * Read an boolean text value line and check separator.
         * The line must contain {@link #TRUE} or {@link #FALSE}.
         *
         * @param   separator   this text must follow the argument
         * @return  resulting boolean value
         * @throws  IOException if any file error occured
         * @throws  IllegalArgumentException if line didn't contain
         *          an boolean text value,.
         *          the separator wasn't found or the
         *          file ended unexpected
         */
        private final boolean readBoolean(final String separator)
                throws IOException {
            argumentLineNumber = getLineNumber();
            argumentBuffer.setLength(0);
            String line = readLine();
            checkNoEndOfFile(line);
            line = line.trim().toLowerCase();
            argumentBuffer.append(line);
            final boolean result;
            if (line.equals(TRUE)) {
                result = true;
            } else if (line.equals(FALSE)) {
                result = false;
            } else {
                throw new IllegalArgumentException("Unexpected value \"" + line
                    + "\" at line " + getLineNumber());
            }
            line = readLine();
            checkSeparator(line, separator);
            return result;
        }


        /**
         * Read an {@link Argument} and check separator.
         * Lines are read till {@link #isSeparator} returns
         * <code>true</code>. These lines should build an
         * {@link Argument} or are equal to {@link #FAIL}.
         *
         * @param   separator   this text must follow the argument
         * @return  created argument or <code>null</code> if
         *          text was equal to {@link #FAIL}
         * @throws  IOException if any file error occured
         * @throws  IllegalArgumentException if argument creation
         *          failed, the separator wasn't found or the
         *          file ended unexpected
         */
        private final Argument readArgumentOrFail(final String separator)
                throws IOException {
            argumentLineNumber = getLineNumber();
            argumentBuffer.setLength(0);
            String line = readLine();
            checkNoEndOfFile(line);
            line = line.trim().toLowerCase();
            argumentBuffer.append(line);
            if (line.equals(FAIL)) {
                line = readLine();
                checkSeparator(line, separator);
                return null;
            } else {
                unReadLine();
                return readArgument(separator);
            }
        }


        /**
         * Check if lowercase and whitespace stripped <code>text</code>
         * is equal to <code>separator</code>.
         *
         * @param   text        current text
         * @param   separator   expected text
         * @throws  IllegalArgumentException    if not equal or
         *          <code>text</code> is <code>null</code>
         */
        private final void checkSeparator(final String text,
                final String separator) throws IllegalArgumentException {
            checkNoEndOfFile(text);
            if (!text.trim().toLowerCase().equals(separator)) {
                throw new IllegalArgumentException("Expected separator was "
                    + separator + " at " + "line " + getLineNumber());
            }
        }

        /**
         * Check if <code>text</code> is <code>null</code>.
         *
         * @param   text    this text is checked
         * @throws  IllegalArgumentException    if <code>text</code> is
         *          <code>null</code>
         */
        private final void checkNoEndOfFile(final String text)
                throws IllegalArgumentException {
            if (text == null) {
                throw new IllegalArgumentException("Unexpected end of file at "
                    + "line " + getLineNumber());
            }
        }


        /**
         * Return argumentNumber.
         *
         * @return  argumentNumber
         */
        private int getArgumentLineNumber() {
            return argumentLineNumber;
        }


        /**
         * Return argument as <code>String</code>.
         *
         * @return  argument
         */
        private String getArgument() {
            return argumentBuffer.toString();
        }


    }




}

